import sys

import fastjsonschema
from encadre import Encadre


class Encadre_Test_App(Encadre):

    def validate_query(self, schema):
        from encadre import request
        fastjsonschema.compile(schema)(request.json)


def serve():
    if len(sys.argv) != 2:
        print("Usage: %s config.ini")
        sys.exit(1)
    Encadre_Test_App(sys.argv[1]).serve()


def dump_routes():
    if len(sys.argv) != 2:
        print("Usage: %s config.ini")
        sys.exit(1)
    Encadre_Test_App(sys.argv[1]).dump_routes()



import sys
import logging
import traceback
from importlib.metadata import entry_points
from decorator import decorator

logger = logging.getLogger(__name__)
ENI = Exception("Not implemented.")


class Cookies():
    def __init__(self, framework):
        self.f = framework

    def get(self, key):
        return self.f.get_cookie(key)

    def set(self, *args, **kwargs):
        return self.f.set_cookie(*args, **kwargs)


class Request():
    def __init__(self, framework):
        self.f = framework
        self._json = None
        self._environ = None
        self._headers = None
        self._args = None
        self._files = None
        self._form = None

    def _get_json(self):
        return self._json or self.f.request_json()

    # Let unit tests set some json
    def _set_json(self, json):
        self._json = json

    json = property(_get_json, _set_json)

    def _get_environ(self):
        return self._environ or self.f.request_environ()

    # Let unit tests set environ
    def _set_environ(self, environ):
        self._environ = environ

    environ = property(_get_environ, _set_environ)

    def _get_headers(self):
        return self._headers or self.f.request_headers()

    # Let unit tests set environ
    def _set_headers(self, headers):
        self._headers = headers

    headers = property(_get_headers, _set_headers)

    def _get_args(self):
        return self._args or self.f.request_args()

    # Let unit tests set environ
    def _set_args(self, args):
        self._args = args

    args = property(_get_args, _set_args)

    def _get_files(self):
        return self._files or self.f.request_files()

    # Let unit tests set files
    def _set_files(self, files):
        self._files = files

    files = property(_get_files, _set_files)

    def _get_form(self):
        return self._form or self.f.request_form()

    # Let unit tests set form
    def _set_form(self, form):
        self._form = form

    form = property(_get_form, _set_form)


def validate(vargs):
    def helper(func, *args, **kwargs):
        if hasattr(args[0].framework, 'validate_query'):
            args[0].framework.validate_query(vargs)
        return func(*args, **kwargs)
    return decorator(helper)


class Framework():

    request_class = Request
    cookies_class = Cookies

    def __init__(self, application, config):
        self.application = application
        self.config = config
        self.first_version = int((self.config or {}).get(application, {}).get(
            'first_version', '0'))
        if self.first_version:
            logger.debug("Config set 'first_version' to %s." %
                         self.first_version)
        self.versioning_use_prefix = (self.config or {}).get(
            application, {}).get('versioning', 'headers').lower() == 'prefix'
        if self.versioning_use_prefix:
            logger.debug("Config set versioning to 'prefix' mode.")
        self.routes = {}
        self.controllers = {}
        self.request = self.request_class(self)
        setattr(sys.modules['encadre'], 'request', self.request)
        self.cookies = self.cookies_class(self)
        setattr(sys.modules['encadre'], 'cookies', self.cookies)

    def add_controller(self, ctrl, methods):
        ctrl_cls, ctrl_name, ctrl_version = ctrl
        self.controllers.setdefault(ctrl_name, {})
        self.controllers[ctrl_name][ctrl_version] = {
            'doc': (ctrl_cls.__doc__ or '').strip(),
            'cls': ctrl_cls,
            'methods': {}}
        for method_name, method, args, verbs in methods:
            self.controllers[ctrl_name][ctrl_version]['methods'].setdefault(
                method_name, {})
            for verb in verbs:
                self.controllers[ctrl_name][ctrl_version]['methods'][
                    method_name][verb] = {
                        'fct': method,
                        'doc': (method.__doc__ or '').strip(),
                        'args': args}

    def setup_application(self, instance):
        self.instance = instance
        self.before_call_chain = []
        if hasattr(instance, 'before_call'):
            self.before_call_chain.append(self.instance_before_call)
        self.after_call_chain = []
        if hasattr(instance, 'after_call'):
            self.after_call_chain.append(self.instance_after_call)
        if hasattr(instance, 'validate_query'):
            self.validate_query = instance.validate_query

    def instance_before_call(self):
        getattr(self.instance, 'before_call')()

    def instance_after_call(self, response):
        getattr(self.instance, 'after_call')()
        return response

    def not_implemented(self):
        raise ENI

    request_json = request_environ = request_headers = \
        request_args = request_files = \
            get_cookie = set_cookie = send_bytes = \
                request_form = not_implemented

    def format_exception(self, e):
        ret = None
        code = 500
        logger.exception(traceback.format_exc())
        if hasattr(self.instance, "format_exception"):
            ret, code = self.instance.format_exception(e)
        else:
            ha = (self.config or {}).get(self.application, {}).get(
                'on_exception', '').lower()
            if ha == 'traceback':
                ret = traceback.format_exc()
            if ha == 'str':
                ret = str(e)
        if ret is None:
            raise e
        return ret, code


class DumpRoutes(Framework):

    def __repr__(self):
        ret = []
        for c in sorted(self.controllers):
            ret.append("* Controller: %s" % (c))
            for v in self.controllers[c]:
                ret.append("** Version: %s: %s" %
                           (v, self.controllers[c][v]['doc']))
                for m in self.controllers[c][v]['methods']:
                    for _v in self.controllers[c][v]['methods'][m]:
                        ret.append(
                            "- [%s] %s%s: %s" %
                            (_v, m,
                             self.controllers[c][v]['methods'][m][_v]['args'],
                             self.controllers[c][v]['methods'][m][_v]['doc']))
        return "\n".join(ret)


def framework_from_config(application, config):
    fw_name = (config.get(application) or {}).get('framework')
    assert fw_name, "No framework configured."
    eps = entry_points()
    app_group = '%s.frameworks' % application
    for group in (app_group, 'encadre.frameworks'):
        group_eps = eps.get(group, [])
        for ep in group_eps:
            if ep.name == fw_name:
                return ep.load()(application, config)
    raise Exception("No framework available")



import unittest
import random
from encadre import Controller, validate


class Hello(Controller):
    """ A very polite controller """

    def index__(self):
        """ Says hello ! """
        return "Hello world"

    def test_index__(self):
        self.assertTrue(self.index__() == "Hello world")

    def config(self):
        """ Dump the config """
        from encadre import config
        return config

    def test_config(self):
        from encadre import config
        k = 'test_%s' % random.randint(0, 1000)
        v = random.randint(0, 1000)
        config.update({k: v})
        self.assertTrue(self.config()[k] == v)

    @validate({
        'title': 'Dummy test',
        'type': 'object',
        'properties': {
            'numbers': {
                'description': "We want an array of numbers",
                'type': 'array',
                'items': {
                    'type': 'integer'
                },
                'minItems': 1
            }
        },
        'required': ['numbers']
    })
    def i_am_validated_POST(self):
        from encadre import request
        return request.json

    def test_i_am_validated_POST(self):
        from encadre import request
        request.json = {'numbers': [1, 2]}
        self.assertTrue(self.i_am_validated_POST() == {
            'numbers': [1, 2]})

    @validate({
        'title': 'Dummy test',
        'type': 'object',
        'properties': {
            'numbers': {
                'description': "We want an array of numbers",
                'type': 'array',
                'items': {
                    'type': 'integer'
                },
                'minItems': 1
            }
        },
        'required': ['numbers']
    })
    def i_will_not_pass_validation_POST(self):
        from encadre import request
        return request.json

    def test_i_will_not_pass_validation_POST(self):
        from encadre import request
        request.json = {'xnumbers': [1, 2]}
        self.i_am_validated_POST()
        self.assertRaises(Exception)


class Hello_2(Controller):
    """ A very polite controller (version 2) """

    def index__(self):
        """ Says hello ! """
        return "Hello world (2)"

    def test_index__(self):
        self.assertTrue(self.index__() == "Hello world (2)")

    def to(self, who):
        """ Says hello to whom ask for """
        return "Hello %s !" % who

    def test_to(self):
        assert True

    def foo(self):
        return "Foo"

    def test_foo(self):
        assert True

    def foo_(self):
        return "Foo with an '_' at the end !"

    def test_foo_(self):
        assert True

    def bar_GET(self):
        return "Bar"

    def test_bar_GET(self):
        assert True

    def baz_POST(self):
        from encadre import request
        return request.json

    def test_baz_POST(self):
        from encadre import request
        k = 'test_%s' % random.randint(0, 1000)
        v = random.randint(0, 1000)
        request.json = {k: v}
        self.assertTrue(self.baz_POST()[k] == v)

    def qux_GET_POST(self):
        return "Qux"

    def test_qux_GET_POST(self):
        assert True

    def args(self):
        from encadre import request
        return request.args.get('qwe', "no 'qwe' arg")

    def test_args(self):
        from encadre import request
        v = 'test_%s' % random.randint(0, 1000)
        request.args['qwe'] = v
        self.assertTrue(self.args() == v)


class Hello_3(Hello_2):
    """ A very polite controller """

    def index__(self):
        """ Says hello ! """
        return "Hello world (3)"

    def test_index__(self):
        self.assertTrue(self.index__() == "Hello world (3)")

    def bar_GET(self):
        """ Better 'bar' """
        return "Better Bar"

    def test_bar_GET(self):
        assert True

    def bar_POST(self):
        """ That one handle POST too ! """
        return "Handle POST too."

    def test_bar_POST(self):
        assert True

    def send_bytes(self):
        return self.framework.send_bytes(
            open("/etc/passwd", "rb"),
            as_attachment=True,
            attachment_filename='etc-passwd.txt',
            mimetype='text/plain')

    def test_send_bytes(self):
        self.assertTrue(b'root' in self.send_bytes())

    def dividebyzero(self):
        return 1 / 0


class StrangeTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass


class Hello_4(Hello_3):

    testcase_cls = StrangeTestCase



import sys
import configparser
import traceback
import logging
import logging.config
from encadre.controllers import Controller, load_controllers
from encadre.framework import Framework, DumpRoutes
from encadre.framework import framework_from_config, validate

logger = logging.getLogger(__name__)


class NotReady:
    def __init__(self, diagnostic):
        self.diagnostic = diagnostic

    def __getattr__(self, attribute):
        raise AttributeError("Cannot access %r: %s" %
                             (attribute, self.diagnostic))

    def __getitem__(self, item):
        raise AttributeError("Cannot find %r: %s" % (item, self.diagnostic))

    def __bool__(self):
        return False


request = cookies = NotReady("Encadre is not setup yet.")


class Encadre():

    config = {}

    def __init__(self, config_filename):
        self.application = getattr(self, 'application',
                                   self.__class__.__name__.lower())
        logger.debug("Starting encadre application '%s'." % self.application)
        self.config_filename = config_filename
        if self.config_filename:
            self.config.update(self.read_config(self.config_filename))
            try:
                logging.config.fileConfig(self.config_filename)
            except Exception:
                print("You'd better configure logging in '%s':\n%s" %
                      (self.config_filename, traceback.format_exc()))
        setattr(sys.modules['encadre'], 'config', self.config)
        self.framework = framework_from_config(self.__class__.__name__.lower(),
                                               self.config)
        self.framework.setup_application(self)
        load_controllers(self.framework, self.application)

    def read_config(self, filename):
        config = {self.__class__.__name__.lower(): {}}
        p = configparser.ConfigParser()
        p.read(filename)
        for n, s in p.items():
            if s:
                config[n] = dict(s)
        return config

    def serve(self):
        assert self.config, "Missing config."
        self.framework.serve()

    def dump_routes(self):
        assert self.config, "Missing config."
        routes = DumpRoutes(self.__class__.__name__.lower(), self.config)
        load_controllers(routes, self.application)
        print(routes)

    def get_wsgi_app(self):
        config = self.read_config(self.config_filename)
        assert config, "Missing config."
        return self.framework.get_wsgi_app()


__all__ = ['Encadre', 'Framework', 'Controller', 'validate']
